@@ -17,13 +17,13 @@
 %                                 April 2007                                  %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2014 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright 1999 ImageMagick Studio LLC, a non-profit organization           %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
 %  obtain a copy of the License at                                            %
 %                                                                             %
-%    http://www.imagemagick.org/script/license.php                            %
+%    https://imagemagick.org/script/license.php                               %
 %                                                                             %
 %  Unless required by applicable law or agreed to in writing, software        %
 %  distributed under the License is distributed on an "AS IS" BASIS,          %
@@ -50,6 +50,7 @@
 #include "magick/list.h"
 #include "magick/magick.h"
 #include "magick/memory_.h"
+#include "magick/option.h"
 #include "magick/pixel-accessor.h"
 #include "magick/property.h"
 #include "magick/quantum-private.h"
@@ -60,6 +61,21 @@
 #include "magick/utility.h"
 #if defined(MAGICKCORE_OPENEXR_DELEGATE)
 #include <ImfCRgbaFile.h>
+#if IMF_VERSION_NUMBER > 1
+#include <OpenEXRConfig.h>
+#endif
+
+/*
+  Typedef declarations.
+*/
+typedef struct _EXRWindowInfo
+{
+  int
+    max_x,
+    max_y,
+    min_x,
+    min_y;
+} EXRWindowInfo;
 
 /*
   Forward declarations.
@@ -134,33 +150,28 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
   const ImfHeader
     *hdr_info;
 
+  EXRWindowInfo
+    data_window,
+    display_window;
+
   Image
     *image;
 
-  ImageInfo
-    *read_info;
-
   ImfInputFile
     *file;
 
   ImfRgba
     *scanline;
 
-  int
-    max_x,
-    max_y,
-    min_x,
-    min_y;
-
   MagickBooleanType
     status;
 
-  register ssize_t
-    x;
-
-  register PixelPacket
+  PixelPacket
     *q;
 
+  size_t
+    columns;
+
   ssize_t
     y;
 
@@ -168,12 +179,12 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
     Open image.
   */
   assert(image_info != (const ImageInfo *) NULL);
-  assert(image_info->signature == MagickSignature);
-  if (image_info->debug != MagickFalse)
+  assert(image_info->signature == MagickCoreSignature);
+  assert(exception != (ExceptionInfo *) NULL);
+  assert(exception->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",
       image_info->filename);
-  assert(exception != (ExceptionInfo *) NULL);
-  assert(exception->signature == MagickSignature);
   image=AcquireImage(image_info);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == MagickFalse)
@@ -181,65 +192,116 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
       image=DestroyImageList(image);
       return((Image *) NULL);
     }
-  read_info=CloneImageInfo(image_info);
-  if (IsPathAccessible(read_info->filename) == MagickFalse)
-    {
-      (void) AcquireUniqueFilename(read_info->filename);
-      (void) ImageToFile(image,read_info->filename,exception);
-    }
-  file=ImfOpenInputFile(read_info->filename);
+  file=ImfOpenInputFile(image->filename);
   if (file == (ImfInputFile *) NULL)
     {
       ThrowFileException(exception,BlobError,"UnableToOpenBlob",
         ImfErrorMessage());
-      if (LocaleCompare(image_info->filename,read_info->filename) != 0)
-        (void) RelinquishUniqueFileResource(read_info->filename);
-      read_info=DestroyImageInfo(read_info);
+      image=DestroyImageList(image);
       return((Image *) NULL);
     }
   hdr_info=ImfInputHeader(file);
-  ImfHeaderDisplayWindow(hdr_info,&min_x,&min_y,&max_x,&max_y);
-  image->columns=max_x-min_x+1UL;
-  image->rows=max_y-min_y+1UL;
+  ImfHeaderDataWindow(hdr_info,&display_window.min_x,&display_window.min_y,
+    &display_window.max_x,&display_window.max_y);
+  image->columns=(size_t) (display_window.max_x-display_window.min_x+1UL);
+  image->rows=(size_t) (display_window.max_y-display_window.min_y+1UL);
   image->matte=MagickTrue;
-  SetImageColorspace(image,RGBColorspace);
+  (void) SetImageColorspace(image,RGBColorspace);
+  image->gamma=1.0;
   if (image_info->ping != MagickFalse)
     {
       (void) ImfCloseInputFile(file);
-      if (LocaleCompare(image_info->filename,read_info->filename) != 0)
-        (void) RelinquishUniqueFileResource(read_info->filename);
-      read_info=DestroyImageInfo(read_info);
       (void) CloseBlob(image);
       return(GetFirstImageInList(image));
     }
-  scanline=(ImfRgba *) AcquireQuantumMemory(image->columns,sizeof(*scanline));
-  if (scanline == (ImfRgba *) NULL)
+  status=SetImageExtent(image,image->columns,image->rows);
+  if (status == MagickFalse)
     {
       (void) ImfCloseInputFile(file);
-      if (LocaleCompare(image_info->filename,read_info->filename) != 0)
-        (void) RelinquishUniqueFileResource(read_info->filename);
-      read_info=DestroyImageInfo(read_info);
-      ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed");
+      (void) CloseBlob(image);
+      InheritException(exception,&image->exception);
+      return(DestroyImageList(image));
+    }
+  ImfHeaderDataWindow(hdr_info,&data_window.min_x,&data_window.min_y,
+    &data_window.max_x,&data_window.max_y);
+  columns=(size_t) (data_window.max_x-data_window.min_x+1UL);
+  if ((display_window.min_x > data_window.max_x) ||
+      (display_window.min_x+(int) image->columns <= data_window.min_x))
+    scanline=(ImfRgba *) NULL;
+  else
+    {
+      scanline=(ImfRgba *) AcquireQuantumMemory(columns,sizeof(*scanline));
+      if (scanline == (ImfRgba *) NULL)
+        {
+          (void) ImfCloseInputFile(file);
+          (void) CloseBlob(image);
+          image=DestroyImageList(image);
+          ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed");
+        }
     }
   for (y=0; y < (ssize_t) image->rows; y++)
   {
+    int
+      yy;
+
+    ssize_t
+      x;
+
     q=QueueAuthenticPixels(image,0,y,image->columns,1,exception);
     if (q == (PixelPacket *) NULL)
       break;
-    ResetMagickMemory(scanline,0,image->columns*sizeof(*scanline));
-    ImfInputSetFrameBuffer(file,scanline-min_x-image->columns*(min_y+y),1,
-      image->columns);
-    ImfInputReadPixels(file,min_y+y,min_y+y);
+    yy=(int) (display_window.min_y+y);
+    if ((yy < data_window.min_y) || (yy > data_window.max_y) ||
+        (scanline == (ImfRgba *) NULL))
+    {
+      for (x=0; x < (ssize_t) image->columns; x++)
+      {
+        SetPixelRed(q,image->background_color.red);
+        SetPixelGreen(q,image->background_color.green);
+        SetPixelBlue(q,image->background_color.blue);
+        SetPixelOpacity(q,image->background_color.opacity);
+        q++;
+      }
+      if (SyncAuthenticPixels(image,exception) == MagickFalse)
+        break;
+      continue;
+    }
+    (void) memset(scanline,0,columns*sizeof(*scanline));
+    if (ImfInputSetFrameBuffer(file,scanline-data_window.min_x-columns*yy,1,
+      columns) == 0)
+      {
+        status=MagickFalse;
+        break;
+      }
+    if (ImfInputReadPixels(file,yy,yy) == 0)
+      {
+        status=MagickFalse;
+        break;
+      }
     for (x=0; x < (ssize_t) image->columns; x++)
     {
-      SetPixelRed(q,ClampToQuantum((MagickRealType) QuantumRange*
-        ImfHalfToFloat(scanline[x].r)));
-      SetPixelGreen(q,ClampToQuantum((MagickRealType) QuantumRange*
-        ImfHalfToFloat(scanline[x].g)));
-      SetPixelBlue(q,ClampToQuantum((MagickRealType) QuantumRange*
-        ImfHalfToFloat(scanline[x].b)));
-      SetPixelAlpha(q,ClampToQuantum((MagickRealType) QuantumRange*
-        ImfHalfToFloat(scanline[x].a)));
+      int
+        xx;
+
+      xx=(int) (display_window.min_x+x-data_window.min_x);
+      if ((xx < 0) || (display_window.min_x+(int) x > data_window.max_x))
+        {
+          SetPixelRed(q,image->background_color.red);
+          SetPixelGreen(q,image->background_color.green);
+          SetPixelBlue(q,image->background_color.blue);
+          SetPixelOpacity(q,image->background_color.opacity);
+        }
+      else
+        {
+          SetPixelRed(q,ClampToQuantum((MagickRealType) QuantumRange*
+            (MagickRealType) ImfHalfToFloat(scanline[xx].r)));
+          SetPixelGreen(q,ClampToQuantum((MagickRealType) QuantumRange*
+            (MagickRealType) ImfHalfToFloat(scanline[xx].g)));
+          SetPixelBlue(q,ClampToQuantum((MagickRealType) QuantumRange*
+            (MagickRealType) ImfHalfToFloat(scanline[xx].b)));
+          SetPixelAlpha(q,ClampToQuantum((MagickRealType) QuantumRange*
+            (MagickRealType) ImfHalfToFloat(scanline[xx].a)));
+        }
       q++;
     }
     if (SyncAuthenticPixels(image,exception) == MagickFalse)
@@ -247,10 +309,12 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
   }
   scanline=(ImfRgba *) RelinquishMagickMemory(scanline);
   (void) ImfCloseInputFile(file);
-  if (LocaleCompare(image_info->filename,read_info->filename) != 0)
-    (void) RelinquishUniqueFileResource(read_info->filename);
-  read_info=DestroyImageInfo(read_info);
-  (void) CloseBlob(image);
+  if (status == MagickFalse)
+    ThrowReaderException(CorruptImageError,"UnableToReadImageData");
+  if (CloseBlob(image) == MagickFalse)
+    status=MagickFalse;
+  if (status == MagickFalse)
+    return(DestroyImageList(image));
   return(GetFirstImageInList(image));
 }
 #endif
@@ -280,19 +344,29 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
 */
 ModuleExport size_t RegisterEXRImage(void)
 {
+  char
+    version[MagickPathExtent];
+
   MagickInfo
     *entry;
 
+  *version='\0';
   entry=SetMagickInfo("EXR");
 #if defined(MAGICKCORE_OPENEXR_DELEGATE)
   entry->decoder=(DecodeImageHandler *) ReadEXRImage;
   entry->encoder=(EncodeImageHandler *) WriteEXRImage;
+#if defined( OPENEXR_PACKAGE_STRING)
+  (void) FormatLocaleString(version,MagickPathExtent,OPENEXR_PACKAGE_STRING);
+#endif
 #endif
   entry->magick=(IsImageFormatHandler *) IsEXR;
+  if (*version != '\0')
+    entry->version=ConstantString(version);
   entry->adjoin=MagickFalse;
   entry->description=ConstantString("High Dynamic-range (HDR)");
   entry->blob_support=MagickFalse;
-  entry->module=ConstantString("EXR");
+  entry->seekable_stream=MagickTrue;
+  entry->magick_module=ConstantString("EXR");
   (void) RegisterMagickInfo(entry);
   return(MagickImageCoderSignature);
 }
@@ -349,6 +423,10 @@ ModuleExport void UnregisterEXRImage(void)
 */
 static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
 {
+  const char
+    *sampling_factor,
+    *value;
+
   ImageInfo
     *write_info;
 
@@ -365,15 +443,17 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
     *scanline;
 
   int
-    compression;
+    channels,
+    compression,
+    factors[3];
 
   MagickBooleanType
     status;
 
-  register const PixelPacket
+  const PixelPacket
     *p;
 
-  register ssize_t
+  ssize_t
     x;
 
   ssize_t
@@ -383,10 +463,10 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
     Open output image file.
   */
   assert(image_info != (const ImageInfo *) NULL);
-  assert(image_info->signature == MagickSignature);
+  assert(image_info->signature == MagickCoreSignature);
   assert(image != (Image *) NULL);
-  assert(image->signature == MagickSignature);
-  if (image->debug != MagickFalse)
+  assert(image->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == MagickFalse)
@@ -416,9 +496,86 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
   if (write_info->compression == B44ACompression)
     compression=IMF_B44A_COMPRESSION;
 #endif
+  channels=0;
+  value=GetImageOption(image_info,"exr:color-type");
+  if (value != (const char *) NULL)
+    {
+      if (LocaleCompare(value,"RGB") == 0)
+        channels=IMF_WRITE_RGB;
+      else if (LocaleCompare(value,"RGBA") == 0)
+        channels=IMF_WRITE_RGBA;
+      else if (LocaleCompare(value,"YC") == 0)
+        channels=IMF_WRITE_YC;
+      else if (LocaleCompare(value,"YCA") == 0)
+        channels=IMF_WRITE_YCA;
+      else if (LocaleCompare(value,"Y") == 0)
+        channels=IMF_WRITE_Y;
+      else if (LocaleCompare(value,"YA") == 0)
+        channels=IMF_WRITE_YA;
+      else if (LocaleCompare(value,"R") == 0)
+        channels=IMF_WRITE_R;
+      else if (LocaleCompare(value,"G") == 0)
+        channels=IMF_WRITE_G;
+      else if (LocaleCompare(value,"B") == 0)
+        channels=IMF_WRITE_B;
+      else if (LocaleCompare(value,"A") == 0)
+        channels=IMF_WRITE_A;
+      else
+        (void) ThrowMagickException(&image->exception,GetMagickModule(),
+          CoderWarning,"ignoring invalid defined exr:color-type","=%s",value);
+   }
+  sampling_factor=(const char *) NULL;
+  factors[0]=0;
+  if (image_info->sampling_factor != (char *) NULL)
+    sampling_factor=image_info->sampling_factor;
+  if (sampling_factor != NULL)
+    {
+      /*
+        Sampling factors, valid values are 1x1 or 2x2.
+      */
+      if (sscanf(sampling_factor,"%d:%d:%d",factors,factors+1,factors+2) == 3)
+        {
+          if ((factors[0] == factors[1]) && (factors[1] == factors[2]))
+            factors[0]=1;
+          else
+            if ((factors[0] == (2*factors[1])) && (factors[2] == 0))
+              factors[0]=2;
+        }
+      else
+        if (sscanf(sampling_factor,"%dx%d",factors,factors+1) == 2)
+          {
+            if (factors[0] != factors[1])
+              factors[0]=0;
+          }
+      if ((factors[0] != 1) && (factors[0] != 2))
+        (void) ThrowMagickException(&image->exception,GetMagickModule(),
+          CoderWarning,"ignoring sampling-factor","=%s",sampling_factor);
+      else if (channels != 0)
+        {
+          /*
+            Cross check given color type and subsampling.
+          */
+          factors[1]=((channels == IMF_WRITE_YCA) ||
+            (channels == IMF_WRITE_YC)) ? 2 : 1;
+          if (factors[0] != factors[1])
+            (void) ThrowMagickException(&image->exception,GetMagickModule(),
+              CoderWarning,"sampling-factor and color type mismatch","=%s",
+              sampling_factor);
+        }
+    }
+  if (channels == 0)
+    {
+      /*
+        If no color type given, select it now.
+      */
+      if (factors[0] == 2)
+        channels=image->matte ? IMF_WRITE_YCA : IMF_WRITE_YC;
+      else
+        channels=image->matte ? IMF_WRITE_RGBA : IMF_WRITE_RGB;
+    }
   ImfHeaderSetCompression(hdr_info,compression);
   ImfHeaderSetLineOrder(hdr_info,IMF_INCREASING_Y);
-  file=ImfOpenOutputFile(write_info->filename,hdr_info,IMF_WRITE_RGBA);
+  file=ImfOpenOutputFile(write_info->filename,hdr_info,channels);
   ImfDeleteHeader(hdr_info);
   if (file == (ImfOutputFile *) NULL)
     {
@@ -436,7 +593,7 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
       write_info=DestroyImageInfo(write_info);
       ThrowWriterException(ResourceLimitError,"MemoryAllocationFailed");
     }
-  ResetMagickMemory(scanline,0,image->columns*sizeof(*scanline));
+  memset(scanline,0,image->columns*sizeof(*scanline));
   for (y=0; y < (ssize_t) image->rows; y++)
   {
     p=GetVirtualPixels(image,0,y,image->columns,1,&image->exception);
@@ -444,16 +601,19 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
       break;
     for (x=0; x < (ssize_t) image->columns; x++)
     {
-      ImfFloatToHalf(QuantumScale*GetPixelRed(p),&half_quantum);
+      ImfFloatToHalf(QuantumScale*(MagickRealType) GetPixelRed(p),
+        &half_quantum);
       scanline[x].r=half_quantum;
-      ImfFloatToHalf(QuantumScale*GetPixelGreen(p),&half_quantum);
+      ImfFloatToHalf(QuantumScale*(MagickRealType) GetPixelGreen(p),
+        &half_quantum);
       scanline[x].g=half_quantum;
-      ImfFloatToHalf(QuantumScale*GetPixelBlue(p),&half_quantum);
+      ImfFloatToHalf(QuantumScale*(MagickRealType) GetPixelBlue(p),
+        &half_quantum);
       scanline[x].b=half_quantum;
       if (image->matte == MagickFalse)
         ImfFloatToHalf(1.0,&half_quantum);
       else
-        ImfFloatToHalf(1.0-QuantumScale*GetPixelOpacity(p),
+        ImfFloatToHalf(1.0-QuantumScale*(MagickRealType) GetPixelOpacity(p),
           &half_quantum);
       scanline[x].a=half_quantum;
       p++;
@@ -466,7 +626,8 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image)
   (void) FileToImage(image,write_info->filename);
   (void) RelinquishUniqueFileResource(write_info->filename);
   write_info=DestroyImageInfo(write_info);
-  (void) CloseBlob(image);
-  return(MagickTrue);
+  if (CloseBlob(image) == MagickFalse)
+    status=MagickFalse;
+  return(status);
 }
 #endif
