@@ -1,11 +1,11 @@
 /*
-  Copyright 1999-2013 ImageMagick Studio LLC, a non-profit organization
+  Copyright @ 1999 ImageMagick Studio LLC, a non-profit organization
   dedicated to making software imaging solutions freely available.
 
-  You may not use this file except in compliance with the License.
+  You may not use this file except in compliance with the License.  You may
   obtain a copy of the License at
 
-    http://www.imagemagick.org/script/license.php
+    https://imagemagick.org/script/license.php
 
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
@@ -15,33 +15,28 @@
 
   MagickCore private methods for internal threading.
 */
-#ifndef _MAGICKCORE_THREAD_PRIVATE_H
-#define _MAGICKCORE_THREAD_PRIVATE_H
+#ifndef MAGICKCORE_THREAD_PRIVATE_H
+#define MAGICKCORE_THREAD_PRIVATE_H
+
+#include "magick/cache.h"
+#include "magick/image-private.h"
+#include "magick/resource_.h"
+#include "magick/thread_.h"
 
 #if defined(__cplusplus) || defined(c_plusplus)
 extern "C" {
 #endif
 
-#include "magick/cache.h"
-#include "magick/resource_.h"
-#include "magick/thread_.h"
-
-/*
-  Single threaded unless workload justifies the threading overhead.
-*/
-#define magick_threads(source,destination,chunk,expression) \
-  num_threads((expression) == 0 ? 1 : \
-    (((chunk) > (16*GetMagickResourceLimit(ThreadResource))) && \
-     (GetImagePixelCacheType(source) != DiskCache)) && \
-    (GetImagePixelCacheType(destination) != DiskCache) ? \
-      GetMagickResourceLimit(ThreadResource) : \
-      GetMagickResourceLimit(ThreadResource) < 2 ? 1 : 2)
-
-#if (__GNUC__ > 3) || ((__GNUC__ == 3) && (__GNUC_MINOR__ > 10))
+#define magick_number_threads(source,destination,chunk,factor) \
+  num_threads(GetMagickNumberThreads(source,destination,chunk,factor))
+#if defined(__clang__) || (__GNUC__ > 3) || ((__GNUC__ == 3) && (__GNUC_MINOR__ > 10))
 #define MagickCachePrefetch(address,mode,locality) \
   __builtin_prefetch(address,mode,locality)
 #else
-#define MagickCachePrefetch(address,mode,locality)
+#define MagickCachePrefetch(address,mode,locality) \
+  magick_unreferenced(address); \
+  magick_unreferenced(mode); \
+  magick_unreferenced(locality);
 #endif
 
 #if defined(MAGICKCORE_THREAD_SUPPORT)
@@ -52,6 +47,33 @@ extern "C" {
   typedef size_t MagickMutexType;
 #endif
 
+static inline int GetMagickNumberThreads(const Image *source,
+  const Image *destination,const size_t chunk,const int factor)
+{
+  const CacheType
+    destination_type = (CacheType) GetImagePixelCacheType(destination),
+    source_type = (CacheType) GetImagePixelCacheType(source);
+
+  size_t
+    max_threads = (size_t) GetMagickResourceLimit(ThreadResource),
+    number_threads = 1,
+    workload_factor = 64UL << factor;
+
+  /*
+    Determine number of threads based on workload.
+  */
+  number_threads=(chunk <= workload_factor) ? 1 :
+    (chunk >= (workload_factor << 6)) ? max_threads :
+    1+(chunk-workload_factor)*(max_threads-1)/(((workload_factor << 6))-1);
+  /*
+    Limit threads for non-memory or non-map cache sources/destinations.
+  */
+  if (((source_type != MemoryCache) && (source_type != MapCache)) ||
+      ((destination_type != MemoryCache) && (destination_type != MapCache)))
+    number_threads=MagickMin(number_threads,4);
+  return((int) number_threads);
+}
+
 static inline MagickThreadType GetMagickThreadId(void)
 {
 #if defined(MAGICKCORE_THREAD_SUPPORT)
@@ -108,9 +130,10 @@ static inline MagickBooleanType IsMagickThreadEqual(const MagickThreadType id)
 static inline size_t GetOpenMPMaximumThreads(void)
 {
 #if defined(MAGICKCORE_OPENMP_SUPPORT)
-  return(omp_get_max_threads());
-#endif
+  return((size_t) omp_get_max_threads());
+#else
   return(1);
+#endif
 }
 
 static inline int GetOpenMPThreadId(void)
@@ -122,21 +145,25 @@ static inline int GetOpenMPThreadId(void)
 #endif
 }
 
+#if defined(MAGICKCORE_OPENMP_SUPPORT)
 static inline void SetOpenMPMaximumThreads(const int threads)
 {
-#if defined(MAGICKCORE_OPENMP_SUPPORT)
   omp_set_num_threads(threads);
 #else
-  (void) threads;
+static inline void SetOpenMPMaximumThreads(const int magick_unused(threads))
+{
+  magick_unreferenced(threads);
 #endif
 }
 
+#if defined(MAGICKCORE_OPENMP_SUPPORT)
 static inline void SetOpenMPNested(const int value)
 {
-#if defined(MAGICKCORE_OPENMP_SUPPORT)
   omp_set_nested(value);
 #else
-  (void) value;
+static inline void SetOpenMPNested(const int magick_unused(value))
+{
+  magick_unreferenced(value);
 #endif
 }
 
